 
#include "socket.h"

struct _uic_client_sock {
	GIOChannel *ch;
	GSocket *sock;
};

/* Create new unblocked socket, connect he to the addr, port,
 * return new GIOChannel
 * If any error - returns NULL and set Gerror
 */
uic_client_sock *
uic_client_sock_new(const gchar *addr, gushort port)
{
	GSocket *sock;
	GSocketAddress *sockaddr;
	GIOChannel *ch;
	GError *err = NULL;

	uic_client_sock *clsock;
	
	sock = g_socket_new(G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_TCP, &err);
	
	if (sock == NULL) {
		uic_errnum = UIC_CLIENT_SOCK_ERR_CREATE;
		uic_set_error(err->message);
		g_clear_error(&err);
		return NULL;
	}
	
	sockaddr = g_inet_socket_address_new(g_inet_address_new_from_string(addr), port);

	if (g_socket_connect(sock, sockaddr, NULL, &err) == FALSE) {
		g_socket_shutdown(sock, TRUE, TRUE, NULL);
		g_object_unref(sock);
		uic_errnum = UIC_CLIENT_SOCK_ERR_CONNECT;
		uic_set_error(err->message);
		g_clear_error(&err);
		return NULL;
	}

	g_socket_set_blocking(sock, FALSE);

#ifdef G_OS_UNIX
	ch = g_io_channel_unix_new(g_socket_get_fd(sock));
#elif defined G_OS_WIN32
	ch = g_io_channel_win32_new_socket(g_socket_get_fd(sock));
#else

#error PRECOMPILE_ERROR: NOT DEFAULT OS

#endif

	if (g_io_channel_set_flags(ch, G_IO_FLAG_IS_READABLE | G_IO_FLAG_IS_WRITEABLE, &err) != G_IO_STATUS_NORMAL) {
		g_io_channel_shutdown(ch, TRUE, NULL);
		g_io_channel_unref(ch);
		g_object_unref(sock);
		uic_errnum = UIC_CLIENT_SOCK_ERR_SET_FLAGS;
		uic_set_error(err->message);
		g_clear_error(&err);
		return NULL;
	}
	
	if (g_io_channel_set_encoding(ch, NULL, &err) != G_IO_STATUS_NORMAL) {
		uic_set_error(err->message);
		g_clear_error(&err);
		g_io_channel_shutdown(ch, TRUE, NULL);
		g_io_channel_unref(ch);
		return NULL;
	}

	clsock = g_malloc(sizeof(uic_client_sock));
	
	clsock->ch = ch;
	clsock->sock = sock;

	return clsock;
}

/*
 * returns 0 on success, -1 on error
 */
int
uic_client_sock_set_blocking(uic_client_sock *sock, gboolean is_blocking)
{
	GIOStatus st;
	GError *err = NULL;

	if (is_blocking == FALSE)
		st = g_io_channel_set_flags(sock->ch, g_io_channel_get_flags(sock->ch) | G_IO_FLAG_NONBLOCK, &err);
	else 
		st = g_io_channel_set_flags(sock->ch, 
		g_io_channel_get_flags(sock->ch) & (G_IO_FLAG_MASK ^ G_IO_FLAG_NONBLOCK), &err);

	if (st != G_IO_STATUS_NORMAL) {
		uic_errnum = UIC_CLIENT_SOCK_ERR_SET_FLAGS;
		uic_set_error(err->message);
		g_clear_error(&err);
		return -1;
	}

	return 0;
}

gint
uic_client_sock_recive(const uic_client_sock *sock, gchar *buf, gsize cnt)
{
	gsize n;
	GIOStatus st;
	GError *err = NULL;

	if (sock->ch->is_readable == FALSE) {
		uic_set_error("isnt_readable");
		return -1;
	}

	st = g_io_channel_read_chars(sock->ch, buf, cnt, &n, &err);

	switch (st) {
	case G_IO_STATUS_NORMAL:
		return n;
		break;
	case G_IO_STATUS_AGAIN:
		uic_errnum = UIC_CLIENT_SOCK_ERR_AGAIN;;
		uic_set_error("no data");
		break;
	default:
		uic_errnum = UIC_CLIENT_SOCK_ERR_UNKNOWN;
		uic_set_error(err->message);
		g_clear_error(&err);
		break;
	}
	return -1;
}

gint
uic_client_sock_send(const uic_client_sock *sock, const gchar *buf, gsize cnt)
{
	gsize n;
	GIOStatus st;
	GError *err = NULL;

	if (sock->ch->is_writeable == FALSE) {
		uic_set_error("isnt_writable");
		return -1;
	}
	
	st = g_io_channel_write_chars(sock->ch, buf, cnt, &n, &err);

	switch (st) {
	case G_IO_STATUS_NORMAL:
		g_io_channel_flush(sock->ch, NULL);
		return n;
		break;
	case G_IO_STATUS_AGAIN:
		uic_errnum = UIC_CLIENT_SOCK_ERR_AGAIN;
		uic_set_error(err->message);
		g_clear_error(&err);
		break;
	default:
		uic_errnum = UIC_CLIENT_SOCK_ERR_UNKNOWN;
		uic_set_error(err->message);
		g_clear_error(&err);
		break;
	}
	return -1;
}

gint
uic_client_sock_close(uic_client_sock *sock)
{
	GIOStatus st;
	GError *err = NULL;

	st = g_io_channel_shutdown(sock->ch, TRUE, &err);
	if (st != G_IO_STATUS_NORMAL) {
		uic_set_error(err->message);
		g_clear_error(&err);
		return -1;
	}
	
	g_io_channel_unref(sock->ch);
	g_object_unref(sock->sock);
	g_free(sock);

	return 0;
}

